library unify.impl.io;

import 'dart:async';
import 'dart:convert';
import 'dart:io';
import 'package:async/async.dart';
import 'package:http/io_client.dart' as inner;
import 'package:http/http.dart' as http;
import 'package:meta/meta.dart';
import 'package:path/path.dart';

import '../../unify.dart';
import 'common/http.dart';

final ContentType _jsonContentType =
    ContentType("application", "json", charset: "utf-8");

HttpClient _createHttpClient() {
  bool trustSelfSigned = true;
  HttpClient httpClient = HttpClient()
    ..badCertificateCallback =
        ((X509Certificate cert, String host, int port) => trustSelfSigned);
  var client = httpClient;
  // client.userAgent = "unify.dart";
  return client;
}

class UnifyClientImpl extends UnifyHttpClient {
  UnifyClientImpl(String email, String password, UnifyAccessToken token,
      UnifyApiEndpoints endpoints,
      {HttpClient client})
      : this.client = client == null ? _createHttpClient() : client,
        super(email, password, token, endpoints);

  final HttpClient client;

  @override
  Future<dynamic> sendHttpRequest(
    String url, {
    bool needsToken: true,
    String extract,
    dynamic body,
    String method,
    Uri baseUrl,
  }) async {
    var uri = baseUrl ?? endpoints.userApiUrl.resolve(url);

    if (endpoints.enableProxyMode) {
      uri = baseUrl ?? uri.replace(queryParameters: {"__unify": "api"});
    }

    Future<HttpClientRequest> futureRequest;

    if (method != null) {
      if (method == "PUT") {
        futureRequest = client.putUrl(uri);
      }
      if (method == "DELETE") {
        futureRequest = client.deleteUrl(uri);
      }
      if (method == "POST") {
        futureRequest = client.postUrl(uri);
      }
      if (method == "GET") {
        futureRequest = client.getUrl(uri);
      }
    } else {
      body == null
          ? futureRequest = client.getUrl(uri)
          : futureRequest = client.postUrl(uri);
    }

    HttpClientRequest request = await futureRequest;
    request.headers.set("User-Agent", "unify.dart");

    if (needsToken) {
      if (!isCurrentTokenValid(true)) {
        await login();
      }
      request.headers.set("Authorization", "Bearer ${token.accessToken}");
    }

    if (body != null) {
      request.headers.contentType = _jsonContentType;
      request.write(const JsonEncoder().convert(body));
      print("Body: $body\n Headers: ${request.headers}");
    }

    // var response = await request.close();
    // var content = await response.transform(utf8.decoder).join();
    final response = await request.close();
    final content =
        await response.cast<List<int>>().transform(utf8.decoder).first;

    if (response.statusCode != 200) {
      throw Exception(
          "Failed to perform action. (Status Code: ${response.statusCode})\n$content");
    }
    var result = const JsonDecoder().convert(content);

    if (result is Map) {
      if (extract != null) {
        return result[extract];
      }
    }

    return result;
  }

  @override
  Future<dynamic> postHttp(
    String url, {
    String extract,
    Map<String, dynamic> body,
    String contentType,
    Uri baseUrl,
    bool raw = false,
  }) async {
    var uri = baseUrl ?? endpoints.userApiUrl.resolve(url);

    inner.IOClient ioClient = inner.IOClient(client);

    var response = await ioClient.post(
      uri,
      headers: {
        HttpHeaders.contentTypeHeader:
            contentType ?? "application/x-www-form-urlencoded",
      },
      body: body,
    );

    if (response.statusCode != 200) {
      throw Exception(
          "Failed to perform action. (Status Code: ${response.statusCode})\n${response.body}");
    }

    if (raw) return response.body;

    var result = const JsonDecoder().convert(response.body);

    return result;
  }

  @override
  Future close() async {
    await client.close();
  }

  @override
  Future<Map<String, dynamic>> uploadFile(
    String url, {
    @required List<Object> fileBits,
    @required String name,
    @required String path,
  }) async {
    var _data = File.fromUri(Uri.parse(path));
    var stream = http.ByteStream(DelegatingStream.typed(_data.openRead()));
    var length = await _data.length();
    var uri = Uri.parse(url);
    var request = http.MultipartRequest("POST", uri);
    var multipartFile = http.MultipartFile(name, stream, length,
        filename: basename(_data.path));
    request.files.add(multipartFile);
    var response = await request.send();

    if (response.statusCode != 200) {
      throw Exception(
          "Failed to perform action. (Status Code: ${response.statusCode})\n${response}");
    }
    var _result = await response.stream.transform(utf8.decoder);
    String _response = await _result.first;
    var result = const JsonDecoder().convert(_response);

    return result;
  }
}
